/*
 * direct disk operation exerciser
 *
 * Copyright (C) 2005 Jens Axboe <axboe@suse.de>
 * Copyright (C) 2006 Jens Axboe <axboe@kernel.dk>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <libaio.h>
#include <getopt.h>
#include <locale.h>

#include <linux/fs.h>

#define MAX_DEPTH	128

static int io_size = 512;		/* bytes */
static int sequential = 1;		/* 0 for random, 1 for sequential */
static unsigned long long range = 2;	/* megabytes */
static char filename[256] = "/dev/sdb";
static int depth = 32;
static int total_run = 15000;
static int display_stats = 0;
static int writing = 0;
static int force = 0;

struct iocb *iocbs[MAX_DEPTH];
struct io_event *events;

struct drand48_data random_state;
static unsigned long long next_block;
static unsigned long long blocks_done;
static unsigned long long nr_blocks;

#define ALIGN(buf)	(((unsigned long) (buf) + (io_size-1)) & ~((io_size-1)))

unsigned long long get_next_offset(void)
{
	unsigned long long block;
	long r;

	if (sequential) {
		if (next_block >= nr_blocks)
			next_block = 0;

		block = next_block;
		next_block += (256 * 1024 / io_size);
		return block;
	}
		
	lrand48_r(&random_state, &r);
	block = (1 + (double) (nr_blocks) * (unsigned long long) r / (RAND_MAX+1.0));
	return block - 1;
}

void init_randomizer(void)
{
	unsigned long seed = 0;

	if (sequential)
		return;

	srand48_r(seed, &random_state);
}

unsigned long time_elapsed(struct timeval *start, struct timeval *end)
{
	int seconds = end->tv_sec - start->tv_sec;
	int useconds = end->tv_usec - start->tv_usec;

	return (1000 * seconds) + (useconds / 1000);
}

void do_io(io_context_t io_ctx, int fd, char *buffer)
{
	struct timeval real_start, start, end;
	int i, ret, cur_depth;
	unsigned long ops, best_ops;
	unsigned long bytes, best_bytes;

	for (i = 0; i < depth; i++) {
		struct iocb *io = iocbs[i];
		char *b = buffer + i * io_size;
		unsigned long long off = get_next_offset();
		unsigned long long offset = off * io_size;

		if (writing)
			io_prep_pwrite(io, fd, b, io_size, offset);
		else
			io_prep_pread(io, fd, b, io_size, offset);
	}

	ret = io_submit(io_ctx, depth, iocbs);
	if (ret < 0) {
		fprintf(stderr, "io_submit: %d\n", ret);
		return;
	}

	cur_depth = ret;
	gettimeofday(&start, NULL);
	gettimeofday(&real_start, NULL);

	best_bytes = best_ops = bytes = ops = 0;
	do {
		struct timespec ts = { .tv_sec = 30, .tv_nsec = 0 };
		unsigned long elapsed;

		ret = io_getevents(io_ctx, 1, cur_depth, events, &ts);
		if (ret < 0) {
			if (ret == EINTR)
				continue;

			fprintf(stderr, "io_getevents: %d\n", ret);
			break;
		} else if (!ret) {	
			fprintf(stderr, "No events completed, depth %d\n", cur_depth);
			cur_depth = 0;
			break;
		}

		cur_depth -= ret;
		ops += (ret * (io_size >> 9));
		bytes += (ret * io_size);
		blocks_done += ret;

		gettimeofday(&end, NULL);

		elapsed = time_elapsed(&start, &end);
		if (elapsed >= 1000) {
			unsigned long kb = bytes >> 10;
			unsigned long o, b;

			o = ops * 1000 / elapsed;
			b = kb * 1000 / elapsed;

			if (display_stats)
				printf("stat: ops/sec = %8lu, KiB/sec = %8lu (depth=%2d)\n", o, b, cur_depth);
			if (o > best_ops)
				best_ops = o;
			if (b > best_bytes)
				best_bytes = b;
				
			bytes = ops = 0;
			gettimeofday(&start, NULL);
		}

		if (total_run) {
			elapsed = time_elapsed(&real_start, &end);
			if (total_run && (elapsed >= total_run)) {
				printf("OPS_SEC=%'lu RATE=%'lu\n",best_ops, best_bytes);
				break;
			}
		}

		for (i = 0; i < ret; i++) {
			struct io_event *ev = events + i;
			struct iocb *io = ev->obj;
			unsigned long long offset = get_next_offset();
			int err;

			offset *= io_size;

			if (writing) {
				io_prep_pwrite(io, fd, io->u.c.buf, io_size, offset);
			} else
				io_prep_pread(io, fd, io->u.c.buf, io_size, offset);

			err = io_submit(io_ctx, 1, &io);
			if (err != 1) {
				fprintf(stderr, "io_submit: %d\n", err);
				goto error;
			}

			cur_depth++;
		}

	} while (1);

error:
	if (cur_depth)
		io_getevents(io_ctx, cur_depth, cur_depth, events, NULL);
}

int parse_options(int argc, char **argv)
{
	static struct option longoptions[] = {
		{ "device",	1,	NULL,	'd' },
		{ "area",	1,	NULL,	'a' },
		{ "random",	0,	NULL,	'r' },
		{ "blocks",	1,	NULL,	'b' },
		{ "queue depth",1,	NULL,	'q' },
		{ "time",	1,	NULL,	't' },
		{ "stats",	1,	NULL,	's' },
		{ "write",	0,	NULL,	'w' },
		{ "force",	0,	NULL,	'w' },
	};
	int c, res;

	while (1) {
		c = getopt_long(argc, argv, "d:a:rb:q:t:swf", longoptions,&res);
		if (c == -1)
			break;

		switch (c) {
			case 'd':
				strncpy(filename, optarg, sizeof(filename)-1);
				break;
			case 'a':
				range = atoi(optarg);
				break;
			case 'r':
				sequential = 0;
				break;
			case 'b':
				io_size = atoi(optarg) * 512;
				break;
			case 'q':
				depth = atoi(optarg);
				if (depth <= 0 || depth > MAX_DEPTH) {
					fprintf(stderr, "queue depth out-of-range %d\n", depth);
					return 1;
				}
				break;
			case 't':
				total_run = atoi(optarg) * 1000;
				if (total_run < 0) {
					fprintf(stderr, "bad runtime\n");
					return 1;
				}
				break;
			case 's':
				display_stats = 1;
				break;
			case 'w':
				writing = 1;
				break;
			case 'f':
				force = 1;
				break;
		}
	}

	printf("Dev=%s, %s, a=%LuMiB, %s, b=%u, q=%u, t=%u\n", filename, writing ? "write" : "read", range, sequential ? "sequential" : "random", io_size / 512, depth, total_run / 1000);
	return 0;
}

unsigned long get_size(int fd)
{
	unsigned long long blocks;
	struct stat sb;

	if (fstat(fd, &sb) == -1) {
		perror("fstat");
		return -1;
	}

	if (S_ISBLK(sb.st_mode)) {
		/*
		 * be a little cautious about writing to a device. if
		 * the user didn't use -f for force, ask for confirmation
		 */
		if (writing && !force) {
			printf("Writing directly to block device, are you sure? CTRL-C to cancel now\n");
			(void) getc(stdin);
		}
		if (ioctl(fd, BLKGETSIZE64, &blocks) == -1) {
			perror("BLKGETSIZE64");
			return -1;
		}
		return blocks >> 20ULL;
	} else if (S_ISREG(sb.st_mode))
		return sb.st_size >> 20ULL;

	fprintf(stderr, "Bad file type (must be file or block device)\n");
	return -1;
}

int main(int argc, char *argv[])
{
	unsigned long mb;
	io_context_t *io_ctx;
	char *orig, *buffer;
	int fd, i;

	if (parse_options(argc, argv))
		return 1;

	if (writing)
		fd = open(filename, O_WRONLY | O_DIRECT);
	else
		fd = open(filename, O_RDONLY | O_DIRECT);

	if (fd == -1) {
		perror("open");
		return 1;
	}

	mb = get_size(fd);
	if (nr_blocks == -1)
		return 1;

	if (!range || (range > mb))
		range = mb;

	nr_blocks = (range << 20ULL) / (unsigned long long) io_size;

	printf("nr_blocks=%Lu (size=%u)\n", nr_blocks, io_size);

	io_ctx = malloc(sizeof(*io_ctx));

	i = io_queue_init(depth, io_ctx);
	if (i) {
		fprintf(stderr, "io_queue_init: %d\n", i);
		return 1;
	}

	init_randomizer();

	orig = malloc((depth + 1) * io_size);
	buffer = (char *) ALIGN(orig);

	for (i = 0; i < depth; i++)
		iocbs[i] = malloc(sizeof(struct iocb));

	setlocale(LC_NUMERIC, "en_US");

	events = malloc(depth * sizeof(struct io_event));

	do_io(*io_ctx, fd, buffer);

	io_destroy(*io_ctx);

	for (i = 0; i < depth; i++)
		free(iocbs[i]);

	free(io_ctx);
	free(events);
	free(orig);
	close(fd);
	return 0;
}
